@@ -1,8 +1,5 @@
 
 from __future__ import absolute_import, print_function
-import sys
-
-from bup.compat import ModuleNotFoundError
 
 from bup import source_info
 try:
@@ -25,7 +22,7 @@ else:
 
 # The ~ in a version is a Debian-style "always less than" marker:
 # https://www.debian.org/doc/debian-policy/ch-controlfields.html#version
-base_version = b'0.32'
+base_version = b'0.33'
 
 version = base_version
 if version.endswith(b'~'):
